/**
 * Copyright 1998-2006, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package org.eclipse.mylar.zest.layouts.algorithms;

import java.util.HashMap;
import java.util.Iterator;

import org.eclipse.mylar.zest.layouts.LayoutStyles;
import org.eclipse.mylar.zest.layouts.dataStructures.DisplayIndependentPoint;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalNode;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship;


/**
 * This algorithm (currently) moves the nodes and arcs in small circles.
 * You can define the total time for the motion, or the number of steps.  You can 
 * also define the angle by which the nodes revolve around their original points.
 * It can be used to highlight a connected graph to make it stand out 
 * from all the other nodes and edges.  
 * 
 * @author Chris Callendar
 */
public class MotionLayoutAlgorithm extends ContinuousLayoutAlgorithm {

	/** The default radius - the distance to move away from the original location. */
	public static final double DEFAULT_RADIUS = 2D;
	
	/** The default sleep time in milliseconds. This determines the speed of the motion.  */
	public static final long DEFAULT_SLEEP = 50L;
	public static final long MIN_SLEEP = 5L;
	
	/** The default total time - this is how long the motion runs for. */
	public static final long DEFAULT_TIME = 1000L;
	
	/** The default number of steps. */
	public static final int DEFAULT_STEPS = 20;
	/** 
	 * The default angle increment.  This is how fast the motion rotates 
	 * in a circle. It is in Pi units. 
	 */
	public static final double DEFAULT_INCREMENT = 0.25D;
	
	// the sleep time - how long to wait between iterations
	private long sleepTime = DEFAULT_SLEEP;
	// the total number of steps
	private int totalSteps = DEFAULT_STEPS;
	// the current iteration/step
	private int step = 0;
	
	// the angle between 0 and 2 (will be multiplied by Pi)
	private double angle = 0;
	// the angle increment between 0 and 2
	private double incr = DEFAULT_INCREMENT;
	// the max angle
	private final double MAX = 2;
	// the radius of the circular motion
	private double radius = DEFAULT_RADIUS;
	
	public MotionLayoutAlgorithm() {
		this(LayoutStyles.NO_LAYOUT_NODE_RESIZING);
	}
	
	public MotionLayoutAlgorithm(int styles) {
		this(styles, DEFAULT_SLEEP, DEFAULT_STEPS);
	}
	
	/**
	 * Constructor.
	 * @param styles the {@link LayoutStyles} for this layout algorithm.  
	 * @param sleepTime the time to sleep between each iteration
	 * @param totalSteps the number of steps to iterate.
	 */
	public MotionLayoutAlgorithm(int styles, long sleepTime, int totalSteps) {
		super(styles);
		setSleepTime(sleepTime);
		setTotalSteps(totalSteps);
	}
	
	/**
	 * Sets the motion radius.  The default is 2, this is the value
	 * of how much a node moves from it's original location.
	 * @param radius the amount to move away from the original location
	 */
	public void setRadius(double radius) {
		this.radius = Math.abs(radius);
	}
	
	public double getRadius() {
		return radius;
	}
	
	/**
	 * The motion is a circular motion.  This value determines how fast the 
	 * motion completes one circle.
	 * @param incr the value between (0, 2)  
	 * 		This value in Pi units. For example, to move 45 degrees each step, 
	 *  	use the default value of 0.25
	 */
	public void setAngleIncrement(double incr) {
		if ((incr > 0) && (incr < MAX)) {
			this.incr = incr;
		}
	}
	
	public double getAngleIncrement() {
		return incr;
	}
	
	/**
	 * The sleep time in milliseconds.
	 * @param sleepTime in milliseconds.
	 */
	public void setSleepTime(long sleepTime) {
		this.sleepTime = Math.max(sleepTime, MIN_SLEEP);
	}

	public long getSleepTime() {
		return sleepTime;
	}
	
	/**
	 * The total number of iterations.
	 * @param totalSteps the number of steps to do the motion
	 */
	public void setTotalSteps(int totalSteps) {
		this.totalSteps = Math.max(0, totalSteps);		
	}
	
	public int getTotalSteps() {
		return totalSteps;
	}
	
	/**
	 * Sets the total run time.  This value is converted into
	 * steps based on the sleep time, and is rounded to the nearest number of 
	 * steps that complete a full circle.
	 * @param time the time in millis
	 */
	public void setTotalRunTime(long totalTime) {
		if (totalTime > 0) {
			int steps = (int)(totalTime / sleepTime);
			// we want one full circle
			double stepsPerCircle = (MAX / incr);
			steps = (int)(Math.round(steps / stepsPerCircle) * stepsPerCircle);
			setTotalSteps(steps);
		}
	}
	
	public long getTotalRunTime() {
		return (totalSteps * sleepTime);
	}

	protected boolean isValidConfiguration(boolean asynchronous, boolean continuous) {
		return true;
	}

	protected void computeOneIteration(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider, double x, double y, double width, double height) {
		// no point in moving!
		if (radius == 0) {
			layoutStopped = true;
			return;
		}
		
		try {
			Thread.sleep(sleepTime);
		} catch (InterruptedException ignore) {			
		}
		
		for (int i = 0; i < entitiesToLayout.length; i++) {
			InternalNode node = entitiesToLayout[i];
			double newX = node.getInternalX() + radius * Math.cos(angle * Math.PI);
			double newY = node.getInternalY() + radius * Math.sin(angle * Math.PI);
			node.setInternalLocation(newX, newY);
		}
		angle = (angle + incr) % MAX;
		
		step++;
		if (step >= totalSteps) {
			layoutStopped = true;
			return;
		}		
	}
	
	protected void applyLayoutInternal(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider, double x, double y, double width, double height) {
		// save the original locations
		HashMap originalLocations = new HashMap(entitiesToLayout.length);
		for (int i = 0; i < entitiesToLayout.length; i++) {
			InternalNode node = entitiesToLayout[i];
			DisplayIndependentPoint dip = new DisplayIndependentPoint(node.getCurrentX(), node.getCurrentX());
			originalLocations.put(node, dip);
		}
		
		// now do the motion
		super.applyLayoutInternal(entitiesToLayout, relationshipsToConsider, x, y, width, height);
		
		// restore the locations
		for (Iterator iter = originalLocations.keySet().iterator(); iter.hasNext(); ) {
			InternalNode node = (InternalNode) iter.next();
			DisplayIndependentPoint dip = (DisplayIndependentPoint)originalLocations.get(node);
			node.setInternalLocation(dip.x, dip.y);
		}
		updateLayoutLocations(entitiesToLayout);
	}

	protected boolean performAnotherNonContinuousIteration() {
		return true;
	}

	protected void preLayoutAlgorithm(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider, double x, double y, double width, double height) {		
		step = 0;
		angle = 0;
	}

	protected void postLayoutAlgorithm(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider) {
		// allow this algorithm to run more than once in asynchronous mode
		started = false;
	}

	protected int getTotalNumberOfLayoutSteps() {
		return totalSteps;
	}

	protected int getCurrentLayoutStep() {
		return step;
	}
		
}
